import base64
import textwrap
from collections import defaultdict
from io import BytesIO
from textwrap import dedent

import PIL.Image as pImage
import numpy as np
from IPython.display import display, HTML


class Image:
    data = None
    title = None

    def __init__(self, rgb=None, filename=None, title=None):
        self.title = title
        if rgb is not None:
            self.data = np.array(rgb).astype(np.uint8)

        self.filename = filename

    @property
    def base64(self):
        if self.data is not None:
            with BytesIO() as buf:
                pImage.fromarray(self.data).save(buf, "png")
                return "data:image/png;base64," + base64.b64encode(buf.getvalue()).decode('utf-8')
        elif self.filename is not None:
            with open(self.filename, "rb") as f:
                encoded = base64.b64encode(f.read()).decode('utf-8')
                return encoded

    @property
    def _html(self):
        if self.title is not None:
            return f'<div><div style="text-align: center">{self.title}</div>' \
                   f'<img style="margin: 0.5em" src="{self.base64 or ""}"/></div>'
        return f'<img style="margin: 0.5em" src="{self.base64 or ""}"/>'


class Text:
    def __init__(self, *args, sep=" ", end="\n", dedent=None):
        t = sep.join(args) + end
        if len(args) == 1:
            dedent = True

        if dedent:
            t = textwrap.dedent(t)
        self._html = f"<div>{t}<div>"


buffer = []


def imshow(img=None, title=None):
    buffer.append(Image(rgb=img, title=title))


def text(*args, sep=" ", end="\n", dedent=None):
    buffer.append(Text(*[a if isinstance(a, str) else str(a) for a in args], sep=sep, end=end, dedent=dedent))


def show(*imgs, titles=None):
    # get parent stack code.

    if titles is None:
        titles = [None] * len(imgs)

    for img, title in zip(imgs, titles):
        buffer.append(Image(rgb=img, title=title))

    # todo: support titles shorter than images.
    string = dedent(
        f"""<div style="display: flex; flex-wrap:wrap; align-items: top; font-family: Times New Roman">{
        ''.join([el._html for el in buffer])}</div>""")

    buffer.clear()
    return display(HTML(string))



def Analysis(fn, prefix="analysis", *ARGS, **KWARGS):
    """
    Analysis wrapper for dumping the results into an analysis folder

    Args:
        fn: the analysis function to be called
        exp: the experiment prefix
        prefix: the prefix for the analysis folder
        *args: positional arguments to be passed to the analysis function
        **kwargs: keyword arguments to be passed to the analysis function

    Returns: returns of the function `fn`
    """
    from ml_logger import RUN

    RUN_DICT = vars(RUN)

    def thunk(*args, **kwargs):
        import time
        import traceback
        from ml_logger import logger

        RUN._update(**RUN_DICT)

        logger.configure(root=RUN.server, prefix=RUN.prefix)
        # logger.job_started(job=dict(job_id=logger.slurm_job_id), host=dict(hostname=logger.hostname), )

        with logger.Prefix(prefix):
            try:
                assert not (args and ARGS), f"can not use position argument at both thunk creation and run.\n" \
                                            f"_args: {args}\n" \
                                            f"ARGS: {ARGS}\n"
                KWARGS.update(**kwargs)

                results = fn(*args, **kwargs)

                with logger.SyncContext():  # Make sure uploaded finished before termination.
                    logger.log_line("========== execution is complete ==========")
                    logger.job_completed()
                    logger.flush()
                time.sleep(3)

            except Exception as e:
                tb = traceback.format_exc()
                with logger.SyncContext():  # Make sure uploaded finished before termination.
                    logger.print(tb, color="red")
                    logger.log_text(tb, filename="traceback.err")
                    # need to add status for ec2/slurm preemption
                    logger.job_errored()
                    logger.flush()
                time.sleep(3)
                raise e

            return results

    return thunk



import os
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from io import BytesIO, StringIO

from ml_logger.requests import SyncRequests
from ml_logger.serdes import serialize, deserialize
from ml_logger.server import LoggingServer
from ml_logger.structs import ALLOWED_TYPES, LogEntry, LogOptions, LoadEntry, RemoveEntry, PingData, GlobEntry, \
    MoveEntry, CopyEntry, MakeVideoEntry, ArchiveEntry, ShellEntry
from requests_futures.sessions import FuturesSession


# noinspection PyPep8Naming
@contextmanager
def _SyncContext(logger, clean=False, max_workers=None):
    """
    Context function for Synchronous network sessions

    :param logger: logger object
    :param clean: remove sync_pool after __exit__()
    :param max_workers: urllib3 max_workers field
    :return:
    """
    old_session = logger.session
    if isinstance(old_session, SyncRequests):
        logger.sync_pool = old_session
    if logger.sync_pool:
        logger.session = logger.sync_pool
    else:
        logger.set_session(False, max_workers or logger.max_workers)
        if not clean:
            logger.sync_pool = logger.session
    try:
        yield
    finally:
        logger.session = old_session


class LogClient:
    local_server: LoggingServer = None
    session = None
    max_workers = None
    # note: used by the context switchers. They do not take parameters, so
    #  only sync/async key is needed to differentiate
    sync_pool = None
    async_pool = None

    def __init__(self, root: str = None, user=None, access_token=None, asynchronous=None, max_workers=None):
        """
        When max_workers is 0, the HTTP requests are synchronous. This allows one to make
        synchronous requests procedurally.

        This is also useful when some logging have to happen before forking subprocesses.
        Mujoco-py for example, would have trouble with forked processes if multiple
        threads are started before forking the subprocesses.

        :param root:
        :param asynchronous: If this is not None, we create a request pool. This way
            we can use the (A)SyncContext call right after construction.
        :param max_workers:
        """
        if asynchronous is not None:
            self.set_session(asynchronous, max_workers)

        if root.startswith("file://"):
            self.local_server = LoggingServer(cwd=root[6:], silent=True, allow_shell=True)
        elif os.path.isabs(root):
            self.local_server = LoggingServer(cwd=root, silent=True, allow_shell=True)
        elif root.startswith('http://') or root.startswith('https://'):
            self.local_server = None  # remove local server to use sessions.
            self.url = root
            self.access_token = access_token
            self.stream_url = os.path.join(root, "stream")
            self.files_url = os.path.join(root, "files")
            self.ping_url = os.path.join(root, "ping")
            self.glob_url = os.path.join(root, "glob")
            self.move_url = os.path.join(root, "move")
            self.copy_url = os.path.join(root, "copy")
            self.make_video_url = os.path.join(root, "make_video")
            self.tar_url = os.path.join(root, "make_archive")
            self.shell_url = os.path.join(root, "shell")
            # when setting sessions the first time, default to use Asynchronous Session.
            if self.session is None:
                asynchronous = True if asynchronous is None else asynchronous
                self.set_session(asynchronous, max_workers)
        else:
            # todo: add https://, and s3://
            raise TypeError('log url need to begin with `/`, `file://`, `http://` or https://.')

    configure = __init__

    def set_session(self, asynchronous, max_workers, proxy=None):
        """
        
        :param asynchronous: bool
        :param max_workers: int for number of workers
        :return: 
        """
        if proxy is None:
            import os
            proxy = os.environ.get('http_proxy', None) or os.environ.get('HTTP_PROXY', None)

        self.max_workers = 10 if max_workers is None else max_workers

        if asynchronous is True:
            self.session = FuturesSession(ThreadPoolExecutor(max_workers=self.max_workers))
        elif asynchronous is False:
            self.session = SyncRequests()

    # noinspection PyPep8Naming
    def SyncContext(self, **kwargs):
        """
        Returns a context object in which the logger logs synchronously.

        :param clean: remove sync_pool after __exit__()
        :param max_workers: urllib3 max_workers field
        :return: context object
        """
        return _SyncContext(self, **kwargs)

    # noinspection PyPep8Naming
    def AsyncContext(self, **kwargs):
        """
        Returns a context object in which the logger logs asynchronously.

        :param clean: remove sync_pool after __exit__()
        :param max_workers: future_request.Session max_workers field
        :return: context object
        """
        return _AsyncContext(self, **kwargs)

    def _get(self, key, dtype, **options):
        if self.local_server:
            return self.local_server.load(key, dtype, **options)
        else:
            json = LoadEntry(key, dtype, **options)._asdict()
            # note: reading stuff from the server is always synchronous via the result call.
            res = self.session.get(self.url, json=json).result()
            # todo: better error handling.
            return deserialize(res.text)

    def _log(self, key, data, dtype, options: LogOptions = None):
        if self.local_server:
            self.local_server.log(key, data, dtype, options)
        else:
            # todo: make the json serialization more robust. Not priority b/c this' client-side.
            json = LogEntry(key, serialize(data), dtype, options)._asdict()
            self.session.post(self.url, json=json)

    # def _shell(self, command, options: ShellOptions=None):
    #     if self.local_server:
    #         self.local_server.exec_shell(command, options)
    #     else:
    #         # todo: make the json serialization more robust. Not priority b/c this' client-side.
    #         json = ShellEntry(command, options)._asdict()
    #         self.session.post(self.shell_url, json=json)

    def stream_download(self, path):
        buf = BytesIO()
        if self.local_server:
            for d in self.local_server.load(path, dtype="byte"):
                buf.write(d)
            buf.seek(0)
            return buf
        else:
            from requests_toolbelt.downloadutils import stream
            # remove leading slash in the path
            full_url = os.path.join(self.files_url, path[1:] if path.startswith('/') else path)

            r = self.session.get(full_url, stream=True)
            stream.stream_response_to_file(r.result(), path=buf)
            buf.seek(0)
            return buf

            # json = LoadEntry(path, "byte")._asdict()
            # # note: reading stuff from the server is always synchronous.
            # #  with (future) sessions, this is forced by the result call.
            # r = self.session.get(self.stream_url, json=json, stream=True)
            # assert stream.stream_response_to_file(r.result(), path=buf) is None
            # buf.seek(0)
            # return buf

    def save_buffer(self, buffer, key):
        # proxy = os.environ.get('HTTP_PROXY')
        # c.setopt(c.PROXY, proxy)
        # logger.print('proxy:', proxy)
        if isinstance(buffer, BytesIO):
            from requests_toolbelt import MultipartEncoder
            encoder = MultipartEncoder({'file': (key, buf), 'canary': true})
            self.session.post(self.url, data=encoder, headers={'Content-Type': encoder.content_type})
        elif isinstance(buffer, StringIO):
            from pycurl import Curl
            c = Curl()
            c.setopt(c.URL, self.url)
            c.setopt(c.TIMEOUT, 3600)
            c.setopt(c.HTTPPOST, [
                ('file', (
                    c.FORM_BUFFER, source_path,
                    c.FORM_BUFFERPTR, buffer.read(),
                    c.FORM_CONTENTTYPE, 'plain/text',
                )),
            ])
            c.perform()
            c.close()

    def save_file(self, source_path, key):
        """

        :param source_path: is relative to the local file system.
        :param key: the key is relative to the current prefix.
        :return:
        """
        if self.local_server:
            if source_path.startswith('/'):
                source_path = "/" + source_path
            return self.local_server.duplicate(source_path, key)
        # proxy = os.environ.get('HTTP_PROXY')
        # c.setopt(c.PROXY, proxy)
        # logger.print('proxy:', proxy)
        from pycurl import Curl, WRITEFUNCTION
        c = Curl()
        # this is to silent response prints
        c.setopt(WRITEFUNCTION, lambda x: None)
        c.setopt(c.URL, self.url)
        c.setopt(c.TIMEOUT, 3600)
        c.setopt(c.HTTPPOST, [
            ('file', (
                c.FORM_FILE, source_path,
                c.FORM_FILENAME, key,
                c.FORM_CONTENTTYPE, 'plain/text',
            )),
        ])
        c.perform()
        c.close()

    def glob(self, query, **kwargs):
        if self.local_server:
            return self.local_server.glob(query, **kwargs)
        else:
            # todo: make the json serialization more robust. Not priority b/c this' client-side.
            json = GlobEntry(query, **kwargs)._asdict()
            res = self.session.post(self.glob_url, json=json).result()
            return res.json()

    def delete(self, key):
        if self.local_server:
            self.local_server.remove(key)
        else:
            # todo: make the json serialization more robust. Not priority b/c this' client-side.
            json = RemoveEntry(key)._asdict()
            self.session.delete(self.url, json=json)

    def move(self, source, target):
        if self.local_server:
            self.local_server.move(source, target)
        else:
            # todo: make the json serialization more robust. Not priority b/c this' client-side.
            json = MoveEntry(source, target)._asdict()
            self.session.post(self.move_url, json=json)

    def duplicate(self, source, target, exists_ok, follow_symlink, symlinks):
        if self.local_server:
            self.local_server.duplicate(source, target, exists_ok, follow_symlink, symlinks)
        else:
            # todo: make the json serialization more robust. Not priority b/c this' client-side.
            json = CopyEntry(source, target, exists_ok, follow_symlink, symlinks)._asdict()
            self.session.post(self.copy_url, json=json)

    def ping(self, exp_key, status, _duplex=True, burn=True):
        # todo: add configuration for early termination
        if self.local_server:
            signals = self.local_server.ping(exp_key, status)
            return deserialize(signals) if _duplex else None
        else:
            # todo: make the json serialization more robust. Not priority b/c this' client-side.
            ping_data = PingData(exp_key, status, burn=burn)._asdict()
            req = self.session.post(self.ping_url, json=ping_data)
            if _duplex:
                response = req.result()
                # note: I wonder if we should raise if the response is non-ok.
                return deserialize(response.text) if response.ok else None

    # send signals to the worker
    def send_signal(self, exp_key, signal=None):
        options = LogOptions(overwrite=True)
        channel = os.path.join(exp_key, "__signal.pkl")
        self._log(channel, signal, dtype="log", options=options)

    # Reads binary data
    def read(self, key, start=None, stop=None):
        return self._get(key, dtype="byte", start=start, stop=stop)

    def read_text(self, key):
        return self._get(key, dtype="text")

    # Reads binary data
    def read_pkl(self, key, start=None, stop=None):
        return self._get(key, dtype="pkl", start=start, stop=stop)

    def read_np(self, key):
        return self._get(key, dtype="np")

    def read_json(self, key):
        return self._get(key, dtype="json")

    def read_h5(self, key):
        return self._get(key, dtype="h5")

    # read_buffer is an alias of read, which returns the buffer.
    read_buffer = read

    # appends data
    def log(self, key, data, **options):
        self._log(key, data, dtype="log", options=LogOptions(**options))

    # appends text
    def log_text(self, key, text, **options):
        self._log(key, text, dtype="text", options=LogOptions(**options))

    # appends yaml
    def log_yaml(self, key, data):
        # does not support appending yet
        self._log(key, data, dtype="yaml")

    # sends out images
    def send_image(self, key, data):
        assert data.dtype in ALLOWED_TYPES, "image data must be one of {}".format(ALLOWED_TYPES)
        self._log(key, data, dtype="image", options=LogOptions(overwrite=True))

    # writes data
    def log_buffer(self, key, buf, **options):
        # defaults to overwrite for binary data.
        self._log(key, buf, dtype="byte", options=LogOptions(**options))
        # self._multipart(key, buf, options=LogOptions(**options))

    # def read_image(self, key):

    def make_video(self, files, key, wd, order, **options):
        if self.local_server:
            return self.local_server.make_video(files=files, key=key, wd=wd, order=order, options=options)
        else:
            # todo: make the json serialization more robust. Not priority b/c this' client-side.
            json = MakeVideoEntry(files=files, key=key, wd=wd, order=order, options=options)._asdict()
            res = self.session.post(self.make_video_url, json=json).result()
            return res.json()['result']

    def make_archive(self, base_name, format="tar", root_dir=None, base_dir=None, **options):
        if self.local_server:
            return self.local_server.make_archive(
                base_name=base_name, format=format, root_dir=root_dir, base_dir=base_dir,
                options=options)
        else:
            # todo: make the json serialization more robust. Not priority b/c this' client-side.
            json = ArchiveEntry(base_name=base_name, root_dir=root_dir, base_dir=base_dir,
                                format=format, options=options)._asdict()
            res = self.session.post(self.tar_url, json=json).result()
            return res.json()['result']

    def shell(self, command, wd, **options):
        if self.local_server:
            return self.local_server.shell(command, wd, options=options)
        else:
            # todo: make the json serialization more robust. Not priority b/c this' client-side.
            json = ShellEntry(command=command, wd=wd, options=options)._asdict()
            res = self.session.post(self.shell_url, json=json).result()
            return res.json()['result']



from json import dumps, loads


class Result:
    def __init__(self, response):
        self.response = response

    def json(self):
        return loads(self.text)

    @property
    def text(self):
        return self.response.data.decode('utf8')

    def ok(self):
        return self.response.ok


class Response:
    def __init__(self, response):
        self.response = response

    def result(self):
        return self.response


class SyncRequests:
    def __init__(self, **kwargs):
        """
        To set the proxy:

        if proxy:
            # conditionally detect proxy setting
            if proxy.startswith("https://"):
                self.session.proxies = {"https": proxy[8:]}
            elif proxy.startswith("http://"):
                self.session.proxies = {"http": proxy[7:]}
            else:
                self.session.proxies = {"http": proxy}

        :param max_workers:
        :param proxy: only one proxy is supported with urllib3.
        """
        from requests import Session

        self.session = Session(**kwargs)

    def get(self, *args, **kwargs):
        _ = self.session.get(*args, **kwargs)
        return Response(_)

    def put(self, *args, **kwargs):
        _ = self.session.put(*args, **kwargs)
        return Response(_)

    def post(self, *args, **kwargs):
        _ = self.session.post(*args, **kwargs)
        return Response(_)

    def delete(self, *args, **kwargs):
        _ = self.session.delete(*args, **kwargs)
        return Response(_)

    def fetch(self, *args, **kwargs):
        _ = self.session.fetch(*args, **kwargs)
        return Response(_)


class xResponse:
    def __init__(self, response):
        self.response = response

    def result(self):
        return Result(self.response)


class xSyncRequests:
    def __init__(self, max_workers=1, proxy=None, **kwargs):
        """
        :param max_workers:
        :param proxy: only one proxy is supported with urllib3.
        """
        import urllib3
        retries = urllib3.Retry(connect=5, read=2, redirect=5)
        if proxy:
            self.pool = urllib3.ProxyManager(proxy_url=proxy, num_pools=max_workers, retries=retries, **kwargs)
        else:
            self.pool = urllib3.PoolManager(num_pools=max_workers, retries=retries, **kwargs)

    def get(self, *args, json, **kwargs):
        _ = self.pool.request('GET', *args, body=dumps(json).encode('utf-8'), **kwargs)
        return xResponse(_)

    def put(self, *args, json, **kwargs):
        _ = self.pool.request('PUT', *args, body=dumps(json).encode('utf-8'), **kwargs)
        return xResponse(_)

    def post(self, *args, json=None, data=None, **kwargs):
        if json is None:
            _ = self.pool.request('POST', *args, body=data, **kwargs)
        else:
            _ = self.pool.request('POST', *args, body=dumps(json).encode('utf-8'), **kwargs)
        return xResponse(_)

    def delete(self, *args, json, **kwargs):
        _ = self.pool.request('DELETE', *args, body=dumps(json).encode('utf-8'), **kwargs)
        return xResponse(_)

    def fetch(self, *args, json, **kwargs):
        _ = self.pool.request('FETCH', *args, body=dumps(json).encode('utf-8'), **kwargs)
        return xResponse(_)



import os
import re
import sys
from os.path import abspath

from params_proto import PrefixProto, Accumulant, Proto, Flag
from termcolor import colored

from .caches.summary_cache import SummaryCache
from .helpers.print_utils import PrintHelper
from .log_client import LogClient
from .ml_logger import USER, ROOT, logger, LOGGER_USER, ML_Logger, pJoin
from .structs import ALLOWED_TYPES


class RUN(PrefixProto, cli_parse=False):
    """The main point of this config object is to provide a way for config functions
    to directly specify the job prefix.

    :param server: the ml-logger server address <host>:<port>
    :param username: the username for the ml-logger server
    :param project: the logging prefix for the project. Default to "scratch"

    :param job_prefix: prefix for this job
    :param job_postfix: postfix for this job
    :param job_counter: a counter. Details see below

    example:
        - job_counter == 0: sets counter to 0
        - job_counter == None: does not use counter in logging prefix.
        - job_counter == True: increment counter by "1" with each job.
    """
    server = ROOT
    username = LOGGER_USER
    project = "scratch"  # default project name

    script_root = Proto(os.getcwd(), help="the root directory of the script. Defautl to current work directory.")
    script_path = None

    now = logger.now()
    prefix = "{username}/{project}/{now:%Y/%m-%d}/{file_stem}/{job_name}"
    job_name = Proto(f"{now:%H.%M.%S}/{{job_counter}}", help="""
        Default to '{now:%H.%M.%S}'. use '{now:%H.%M.%S}/{job_counter}'
        for multiple launches. You can do so by setting:

        ```python
        RUN.job_name += "/{job_counter}"

        for params in sweep:
           thunk = instr(main)
           jaynes.run(thun)
        jaynes.listen()
        ```
        """)
    job_counter = Accumulant(0, help="Default to 0. Use True to increment by 1.")

    resume = Proto(True, help="whether starting the run from scratch, or "
                              "resume previous checkpoints")
    readme = None

    debug = Flag(default="pydevd" in sys.modules, help="set to True automatically for pyCharm")
    CUDA_VISIBLE_DEVICES: str = None

    # noinspection PyMissingConstructor
    @classmethod
    def __new__(cls, deps=None, script_path=None, __count=True, **kwargs):
        from ml_logger import logger

        sr = cls.script_root
        script_root_depth = str(sr.value if isinstance(sr, Proto) else sr).split('/').__len__()
        script_truncated = logger.truncate(os.path.abspath(script_path), depth=script_root_depth)

        file_stem = logger.stem(script_truncated)

        if __count and (isinstance(cls.job_counter, int) or isinstance(cls.job_counter, float)):
            cls.job_counter += 1

        data = vars(cls)
        data.update(kwargs)
        while "{" in data['prefix']:
            data = {k: v.format(file_stem=file_stem, **data) if isinstance(v, str) else v for k, v in data.items()}

        return data['prefix'], data['job_name'], file_stem


if __name__ == '__main__':
    RUN(job_counter=None)
    assert RUN.job_counter is None

    RUN()
    assert RUN.job_counter == 0

    RUN()
    assert RUN.job_counter == 1

    RUN(job_counter=True)
    assert RUN.job_counter == 2

    RUN(job_counter=0)
    assert RUN.job_counter == 0

    RUN(job_counter=10)
    assert RUN.job_counter == 10


def instr(fn, *ARGS, __file=False, __create_job=True, __count=True, __diff=True, __silent=False, **KWARGS):
    """
    Instrumentation thunk factory for configuring the logger.

    :param fn: function to be called
    :param ARGS: position arguments for the call
    :param __file: str, console mode, by-pass file related logging
    :param __create_job: bool, default to True, set to False when you are relaunching
    :param __count: bool, increase the RUN.job_counter by 1 on each `instr` call.
    :param __diff: bool, default to true, when off skips the git diff step.
    :param __silent: bool, do not print
    :param KWARGS: keyword arguments for the call
    :return: a thunk that can be called without parameters
    """
    import inspect
    from termcolor import cprint
    from ml_logger import logger, pJoin

    if __file:
        caller_script = pJoin(os.getcwd(), __file)
    else:
        launch_module = inspect.getmodule(inspect.stack()[1][0])
        __file = launch_module.__file__
        caller_script = abspath(__file)

    # need to set the deps
    # note: for scripts in the `plan2vec` module this also works -- b/c we truncate fixed depth.
    PREFIX, JOB_NAME, FILE_STEM = RUN(script_path=caller_script, _RUN__count=__count)

    RUN_DICT = vars(RUN)
    RUN_DICT['prefix'] = PREFIX
    RUN_DICT['job_name'] = JOB_NAME
    RUN_DICT['file_stem'] = FILE_STEM

    # todo: there should be a better way to log these.
    # todo: we shouldn't need to log to the same directory, and the directory for the run shouldn't be fixed.
    logger.configure(root=RUN.server, prefix=PREFIX, asynchronous=False,  # use sync logger
                     max_workers=4)
    if __create_job:
        # # this is debatable
        # if RUN.restart:
        #     with logger.Sync():
        #         logger.remove(".")
        logger.upload_file(caller_script)

        # the tension is in between creation vs run. Code snapshot are shared, but runs need to be unique.
        logger.job_created(
            job=dict(script_path=RUN.script_path, script_root=RUN.script_root,
                     counter=RUN.job_counter, prefix=RUN.prefix, name=RUN.job_name),
            revision=logger.rev_info(),
            fn=logger.fn_info(fn),
            args=ARGS,
            kwargs=KWARGS)

        if __diff:
            logger.print('taking diff, if this step takes too long, check if your uncommitted changes are '
                         'too large.', color="green")
            logger.diff()

        if RUN.readme:
            logger.log_text(RUN.readme, "README.md", dedent=True)

    try:
        import jaynes  # now set the job name to prefix

        if jaynes.Jaynes.mode not in [False, 'local']:
            assert jaynes.Jaynes.launcher, "Make sure you call jaynes.config first."

            # gcp requires lower-case and less than 60 characters
            # fixme: change all non alpha-numeric and non "-" characters to "-".
            launch_name = USER + "-" + PREFIX[-61 + len(USER):].replace('/', '-').replace('_', '-') \
                .replace('.', '-').lower()
            runner_name = launch_name

            if RUN.CUDA_VISIBLE_DEVICES is not None:
                envs_str = jaynes.Jaynes.runner_config[1]['envs']

                if re.search(r"CUDA_VISIBLE_DEVICES=([A-z0-9\-,]*)", envs_str):
                    envs_ext = re.sub(r"CUDA_VISIBLE_DEVICES=([A-z0-9\-,]*)",
                                      f'CUDA_VISIBLE_DEVICES={RUN.CUDA_VISIBLE_DEVICES}',
                                      envs_str)
                else:
                    envs_ext =  f'{envs_str} CUDA_VISIBLE_DEVICES={RUN.CUDA_VISIBLE_DEVICES}'

                # f" CUDA_VISIBLE_DEVICES={RUN.CUDA_VISIBLE_DEVICES}"
                jaynes.Jaynes.config(jaynes.Jaynes.mode,
                                     launch={'name': launch_name},
                                     runner={'name': runner_name, 'envs': envs_ext})
            else:
                jaynes.Jaynes.config(jaynes.Jaynes.mode,
                                     launch={'name': launch_name},
                                     runner={'name': runner_name})

            del jaynes
            if not __file:
                cprint(f'Set up job name', "green")

    except ModuleNotFoundError:
        pass

    del logger

    def thunk(*args, **kwargs):
        import traceback
        from ml_logger import logger

        RUN._update(**RUN_DICT)
        if RUN.CUDA_VISIBLE_DEVICES is not None:
            visibility = os.environ.get('CUDA_VISIBLE_DEVICES', None)
            assert RUN.CUDA_VISIBLE_DEVICES == visibility, \
                f"CUDA_VISIBLE_DEVICES={visibility}. Expected {RUN.CUDA_VISIBLE_DEVICES}"

        logger.configure(root=RUN.server, prefix=PREFIX, max_workers=10)
        # todo logger.job_id is specific to slurm
        logger.job_started(job=dict(job_id=logger.slurm_job_id), host=dict(hostname=logger.hostname), )

        import time
        try:
            assert not (args and ARGS), f"can not use position argument at both thunk creation and run.\n" \
                                        f"_args: {args}\n" \
                                        f"ARGS: {ARGS}\n"

            _KWARGS = {**KWARGS}
            _KWARGS.update(**kwargs)

            results = fn(*(args or ARGS), **_KWARGS)

            with logger.SyncContext():  # Make sure uploaded finished before termination.
                logger.log_line("========== execution is complete ==========")
                logger.job_completed()
                logger.flush()
            time.sleep(3)

        except Exception as e:
            tb = traceback.format_exc()
            with logger.SyncContext():  # Make sure uploaded finished before termination.
                logger.print(tb, color="red")
                logger.log_text(tb, filename="traceback.err")
                # need to add status for ec2/slurm preemption
                logger.job_errored()
                logger.flush()
            time.sleep(3)
            raise e

        return results

    return thunk


def is_stale(time_str, threshold=5.):
    """Helper function for time lapsed

    :param time_str: the runTime string from logger job.runTime
    :param threshold: in minutes
    :return: boolean value whether time has passed.
    """
    if time_str is None:
        return False

    from datetime import datetime

    time_elapsed = logger.utcnow() - datetime.strptime(time_str, "%Y-%m-%d %H:%M:%S.%f")
    minutes_since = time_elapsed.total_seconds() / 60.
    return minutes_since > threshold


def needs_relaunch(prefix, stale_limit=5., silent=False, not_exist_ok=False, *args, **kwargs):
    """

    :param prefix: the prefix for the experiment entry.
    :param stale_limit: the stale limit in minutes. Shared for all job status.
    :param silent: default to False. When truful, silences the printout.
    :param not_exist_ok: default to False b/c this error message is helpful.
    :param args: extended parameters for printing.
    :return: bool needs relaunch
    """
    # make sure that the prefix context is set to absolute path
    if not prefix.startswith('/'):
        prefix = "/" + prefix

    with logger.Prefix(prefix):
        status, request_time, request_id, region, run_time, start_time, create_time = \
            logger.read_params('job.status', 'job.requestTime', 'job.request_id', 'job.region', 'job.runTime',
                               'job.startTime', 'job.createTime', default=None, not_exist_ok=not_exist_ok)
    stale = relaunch = False
    if status is None:
        s = colored("not exist", 'red'), f"https://app.dash.ml{prefix}"
        relaunch = True
    elif status == "completed":
        s = colored(status, 'green'), f"https://app.dash.ml{prefix}"
    elif status == "created":
        s = colored(status, 'yellow'), f"https://app.dash.ml{prefix}"
        stale = is_stale(create_time, stale_limit)
    elif status == "requested":
        s = colored(status, 'yellow'), f"https://app.dash.ml{prefix}"
        stale = is_stale(request_time, stale_limit)
    elif status == "started":
        s = colored(status, 'yellow'), f"https://app.dash.ml{prefix}"
        stale = is_stale(start_time, stale_limit)
    elif status == "errored":
        s = colored(status, 'red'), f"https://app.dash.ml{prefix}"
        relaunch = True
    elif status == "running":
        s = colored(status, 'yellow'), f"https://app.dash.ml{prefix}"
        stale = is_stale(run_time, stale_limit)
    else:
        s = status,

    if not silent:
        print(*s, colored("is stale", "red") if stale else None, *[a for a in args if a], **kwargs)

    return stale or relaunch


def memoize(f_or_ctx):
    if not callable(f_or_ctx):
        ctx = f_or_ctx

        import hashlib
        hash = hashlib.md5(ctx)

        def inner(f):
            c_path = f"{hash}.{f.__module__}.{f.__name__}.pkl"
            l = ML_Logger(".cache", root=os.getcwd())
            cache = l.load_pkl(c_path)
            memo = cache[0] if cache else {}

            def wrapper(*args, **kwargs):
                key = (*args, *kwargs.keys(), *kwargs.values())
                if key not in memo:
                    memo[key] = f(*args, **kwargs)
                    l.save_pkl(memo, c_path)
                return memo[key]

            return wrapper

        return inner

    f = f_or_ctx
    c_path = f"{f.__module__}.{f.__name__}.pkl"

    l = ML_Logger(".cache", root=os.getcwd())
    cache = l.load_pkl(c_path)
    memo = cache[0] if cache else {}

    def wrapper(*args, **kwargs):
        key = (*args, *kwargs.keys(), *kwargs.values())
        if key not in memo:
            memo[key] = f(*args, **kwargs)
            l.save_pkl(memo, c_path)
        return memo[key]

    return wrapper



from .caches.summary_cache import SummaryCache
from .helpers.print_utils import PrintHelper
from .log_client import LogClient
from .ml_logger import USER, ROOT, logger, LOGGER_USER, ML_Logger, pJoin
from .structs import ALLOWED_TYPES



from collections import defaultdict, deque
import numpy as np
from scipy import stats
import more_itertools as mit

AVERAGE_MODES = ["rolling", 'tiled']
STAT_MODES = ["mean", "min_max", "std_dev", "quantile", "histogram"]


class SummaryCache:
    def __init__(self, mode="tiled", default_stats="mean", window: int = None):
        """

        :param mode:
        :param window: the rolling average window under rolling mode
        """
        self.mode = mode
        if mode == "rolling":
            self.data = defaultdict(lambda: deque(maxlen=window))
        elif mode == "tiled":
            assert window is None, "shouldn't specify window under tiled mode"
            self.data = defaultdict(list)
        else:
            raise KeyError(f'mode `{mode}` is not supported. Allowed modes: {AVERAGE_MODES}.')

        self.default_stats = default_stats

    def __bool__(self):
        """returns True if any of the values is non-empty."""
        for v in self.data.values():
            if bool(v):
                return True
        return False

    def __repr__(self):
        s = "SummaryCache:"
        for k, v in self.data.items():
            try:
                s += f"  {k}: Shape{v[0].shape}[:{len(v)}]\n"
            except:
                s += f"  {k}: {v[:2]}[:{len(v)}]\n"
        return s

    def __contains__(self, item):
        return item in self.data

    def store(self, metrics=None, **key_values):
        """
        Store the metric data for making the summary later. This allows the logging/saving
        of training metrics asynchronously from the logging.

        :param metrics: (mapping) a mapping of metrics key/values. Will be destructured and appended
                        to the data store one key/value at a time,
        :param ** key_values: (Any) key/value arguments, each being a metric key / metric value pair.
        :return: None
        """
        keys = []
        if metrics:
            key_values.update(metrics)
        for k, v in key_values.items():
            self.data[k].append(v)
            keys.append(k)
        return keys

    def summarize(self, force_clear=None, key_stats=None, **_key_stats):
        """
        | summarizes the statistics, and clears the data store if
        |     1. `force_clear` is set to True
        |       OR
        |     2. the cache is under `tiled` mode.
        |
        | Note that this method is NOT Idempotent. Clears the data store if not using rolling average.
        | Because of this, summarize does not support explicit mode of `get_stats`. The summary
        | of all metric fields are returns when calling this method.
        |
        :param force_clear: (bool) forces clear the data store
        :param key_stats: (dict) a dictionary for the key and the statistic modes to be returned.
        :param _key_stats: (**) key value pairs, as a short hand for the key_modes dictionary.
        :return: dictionary of the keys and the statistics requested.
        """
        summary = self.summary_stats(key_stats=key_stats, **_key_stats)
        if force_clear or self.mode == "tiled":
            self.clear()
        return summary

    def clear(self):
        self.data.clear()

    def peek(self, *keys, len=5):
        """
        returns truncated version of the cached metrics, useful for taking a peek of
        what is saved in the dataset at the moment.

        :param keys:
        :param len:
        :return:
        """
        return {k: self.data[k][-len if len else None:] for k in (keys or self.data.keys()) if self.data[k] != []}

    def __getitem__(self, item):
        return self.get(item)

    def get(self, key, default=None):
        """
        Return the summary statistics or default

        :param key: the key for the metric
        :param default: None
        :return:
        """
        if key in self.data:
            cached = self.data[key]
            return np.array(cached)
        else:
            return default

    def pop(self, key, default=None):
        """
        Return the summary staatistics or default, and remove the key from the dictionary.

        example:
            avg_loss, = summary_cache.get('loss', stats='mean')

        :param key:
        :param default:
        :return:
        """
        if key in self.data:
            cached = self.data.pop(key)
            return np.array(cached)
        else:
            return default

    # note: is idempotent
    def summary_stats(self, *only_keys, key_stats=None, explicit=None, stats=None, **_key_stats):
        """
        Returns the metrics as a dictionary containing key / value pairs
        of the statistics of the metrics.

        OR

        Returns statistics that are queried.

        Note that this method is idempotent. AKA you can call multiple times without
        affecting what is stored in the data object.
        
        When a few key strings are passed in explicitly, ONLY those keys
        plus those specified in the keyword arguments in **key_modes are
        returned.
        
        To enable explicit mode without specifying *only_keys, set
        `get_only` to True
        
        # Modes for the Statistics:
        ===================
        
        key_mode would be one of:
          - mean:
          - min_max:
          - std_dev:
          - quantile:
          - histogram(bins=10):
        
        :param * only_keys: list of key strings to return explicitly
        :param key_stats: a dictionary for the key and the statistic modes to be returned.
        :param explicit: boolean flag, when true only keys explicitly specified would be returned
        :param stats: the default stats mode for all metrics
        :param ** _key_stats: key value pairs, as a short hand for the key_modes dictionary.
        :return: dictionary of the keys and the statistics requested
        # :return: dictionary of the keys and the statistics requested, or the value itself if only
        #         one value is specified.
        """
        explicit = explicit or len(only_keys) > 0
        key_stats = key_stats or dict()
        key_stats.update(_key_stats)
        metrics = dict()
        keys = {*only_keys, *key_stats.keys()} if explicit else self.data.keys()
        for k in keys:
            stats_type = key_stats.get(k, stats or self.default_stats)
            if k not in self.data:
                continue
            else:
                # this converts the None to np.nan that can be properly handled.
                d = np.array(flatten(self.data[k]))
                d = d[~np.isnan(d)]
                if len(d) == 0:
                    continue
                if "sum" in stats_type:
                    metrics[k + "/sum"] = d.sum()
                if "max" in stats_type:
                    metrics[k + "/max"] = d.max()
                if "min" in stats_type:
                    metrics[k + "/min"] = d.min()
                if "mean" in stats_type:
                    metrics[k + "/mean"] = d.mean()
                if "min_max" == stats_type:
                    metrics[k + "/mean"] = d.mean()
                if "std" in stats_type:
                    metrics[k + "/stddev"] = d.std()
                    metrics[k + "/mean"] = d.mean()
                    mode = mit.first(stats.mode(d, axis=None)[0], np.nan)
                    metrics[k + "/mode"] = mode
                if "quantile" in stats_type:
                    metrics[k + "/0"] = np.percentile(d, 0)
                    metrics[k + "/25"] = np.percentile(d, 25)
                    metrics[k + "/mean"] = np.percentile(d, 50)
                    metrics[k + "/75"] = np.percentile(d, 75)
                    metrics[k + "/100"] = np.percentile(d, 100)
                if "histograph" in stats_type:
                    # note: need make bin number configurable
                    metrics[k + "/hist"], metrics[k + "/divs"] = np.histogram(d, bins=10)

        return metrics


def to_float(item):
    if item is None:
        return float('nan')
    return float(item)


def flatten(array):
    """
    Returns a flattened nested arrays. Works with numpy containers with non-equal-length children.

    we don't use np.flatten, b/c sometimes you have to iterate through each child.

    Cast None to float('nan') values.

    :param array:
    :return:
    """
    r = []
    try:
        for item in array:
            r += flatten(item)
    except TypeError:  # Iterating through a 0-d array
        r += [to_float(array)]
    return r



class DefaultSet(set):
    def __init__(self, *items):
        """DefaultSet, a set that `reset()`s to the original elements:

        Usage:

        .. code-block:: python

            default_set = DefaultSet("first_default_item", "second_item")
            default_set.reset() # => DefaultSet({"first_default_item", "second_item"})

        :param * items: list of items (as positional arguments) as default for the set.
        """
        super().__init__()
        self._default_set = items

    def reset(self):
        self.clear()
        self.update(self._default_set)

    # def __repr__(self):
    #     return f"{self._default_set}"


if __name__ == "__main__":
    s = DefaultSet('this', '0')

    s.reset()
    assert "this" in s, "should contain original elements"

    s.add('yo')
    assert 'yo' in s, "new key"

    s.add('__yo')
    assert '__yo' in s, "new key"

    s.add('__timestamp')
    assert '__timestamp' in s, "timestamp"

    assert 'not' not in s, "not in s"

    s.clear()
    assert not s, "s should be empty after clear()"
    print('all tests pass!')



class KeyValueCache:
    def __init__(self):
        """ key-value cache for the logger. This one allows one to overwrite old values with new ones. """
        self.data = dict()

    def peek(self):
        return self.data

    def __bool__(self):
        """Returns Trueness of the data dictionary."""
        return bool(self.data)

    def update(self, key_values=None, **kwargs):
        if key_values is not None:
            self.data.update(key_values)
        for k, v in kwargs.items():
            self.data[k].append(v)

    def pop_all(self):
        data = self.data.copy()
        self.data.clear()
        return data

    def clear(self):
        self.data.clear()



import io
import pickle
from random import randint


class Whatever(pickle.Unpickler):
    def __init__(self, *_, **__):
        super().__init__(*_, **__)

    @staticmethod
    def cls_factory(module, name):

        class Unavailable:
            __dict__ = {}

            def __init__(self, *args, **kwargs):
                print(args, kwargs)
                self.__d = dict()

            def __setitem__(self, key, value):
                self.__d[key] = value

            def __getitem__(self, item):
                return self.__d[item]

            def __call__(self, *args, **kwargs):
                print(args, kwargs)

            def __repr__(self):
                return f"<class '{module}.{name}'>"

        return Unavailable

    def find_class(self, module, name):
        try:
            return super().find_class(module, name)
        except:
            return self.cls_factory(module, name)


def load_from_pickle(path='parameters.pkl', **__):
    with open(path, 'rb') as f:
        yield from load_from_pickle_file(f, **__)


def load_from_pickle_file(file, **__):
    while True:
        try:
            yield Whatever(file, **__).load()
        except EOFError:
            break

def load_from_torch_file(file, **kwargs):
    import torch
    while True:
        try:
            yield torch.load(file, **kwargs)
        except EOFError:
            break


def load_from_jsonl_file(file: io.StringIO, **__):
    import json
    for l in file.readlines():
        trimmed = l.strip()
        if trimmed.startswith(b'//'):
            # support comment line
            pass
        else:
            yield json.loads(l)


def regularize_for_json(obj):
    from typing import Sequence
    from types import FunctionType
    import numpy as np
    if isinstance(obj, dict):
        return {k: regularize_for_json(v) for k, v in obj.items()}
    # elif isinstance(obj, tuple):
    #     return (regularize_for_json(i) for i in obj)
    # elif isinstance(obj, list):
    #     return [regularize_for_json(i) for i in obj]
    elif isinstance(obj, int):
        return obj
    elif isinstance(obj, float):
        return obj
    elif isinstance(obj, str):
        return obj
    elif isinstance(obj, np.ndarray) or isinstance(obj, np.ma.MaskedArray):
        return obj.tolist()
    elif isinstance(obj, Sequence):
        return [regularize_for_json(i) for i in obj]
    elif isinstance(obj, FunctionType):
        # todo: use $type syntax:
        #   $type: lambda
        #   $type: function
        return repr(obj)
    try:
        # todo: use $type syntax: $type: class
        return {k: regularize_for_json(v) for k, v in vars(obj).items()}
    except:
        return repr(obj)


def load_from_file(path):
    with open(path, 'rb') as f:
        while True:
            blob = f.read()
            if blob == b'':
                break
            yield blob


def sample(stream, k):
    """

    :param stream:
    :param k: the reservoir size
    :return:
    """

    reservoir = []
    for count, d in enumerate(stream):
        if count < k:
            reservoir.append((count, d))
        else:
            ind = randint(0, count)
            if ind < k:
                reservoir[ind] = (count, d)
    for i, d in sorted(reservoir, key=lambda t: t[0]):
        yield d


def load_pickle_as_dataframe(path='data.pkl', k=None):
    import pandas
    if k:
        d = pandas.DataFrame([_ for _ in sample(load_from_pickle(path), k)])
    else:
        d = pandas.DataFrame([_ for _ in load_from_pickle(path)])
    return d


if __name__ == "__main__":
    sampled = sample(range(100), 1)
    print(list(sampled))
    sampled = sample(range(100), 2)
    print(list(sampled))
    sampled = sample(range(100), 3)
    print(list(sampled))
    sampled = sample(range(100), 30)
    print(list(sampled))



from collections import namedtuple
from typing import NamedTuple, Any


class LogOptions(NamedTuple):
    overwrite: bool = None
    write_mode: str = None


class LogEntry(NamedTuple):
    key: str
    data: Any
    type: str
    options: LogOptions = None


class LoadEntry(NamedTuple):
    key: str
    type: str
    start: Any = None
    stop: Any = None


RemoveEntry = namedtuple("RemoveEntry", ['key'])


class GlobEntry(NamedTuple):
    query: str
    wd: str = "."
    recursive: bool = True
    start: Any = None
    stop: Any = None


class MoveEntry(NamedTuple):
    source: str
    to: str


class CopyEntry(NamedTuple):
    source: str
    to: str
    # for files
    exists_ok: bool
    follow_symlink: bool
    # for directories
    symlinks: bool


class PingData(NamedTuple):
    exp_key: str
    status: Any
    burn: bool = False


class MakeVideoEntry(NamedTuple):
    files: str
    key: str
    wd: str = "."
    order: str = None
    options: Any = None


class ArchiveEntry(NamedTuple):
    base_name: str
    root_dir: str
    base_dir: str
    format: str = "tar"
    options: Any = None


class ShellEntry(NamedTuple):
    command: str
    wd: str = "."
    options: Any = None


Signal = namedtuple("Signal", ['exp_key', 'signal'])

import numpy as np

ALLOWED_TYPES = (np.uint8, np.uint16)  # ONLY uint8 and uint16 is supported.
